@@ -58,43 +58,145 @@ from time import time
 class Thread(threading.Thread):
     def __init__(self, group=None, target=None, name=None, args=(), kwargs=None, max_results=0, daemon=None):
         self.results = Queue(max_results)
-        threading.Thread.__init__(self, group, target, name, (self.results,) + args, kwargs, daemon=daemon)
+        threading.Thread.__init__(self, group, target, name, args, kwargs, daemon=daemon)
+
+    def run(self):
+        """Method representing the thread's activity.
+
+        You may override this method in a subclass. The standard run() method
+        invokes the callable object passed to the object's constructor as the
+        target argument, if any, with sequential and keyword arguments taken
+        from the args and kwargs arguments, respectively.
+
+        """
+        try:
+            if self._target:
+                self.results.put(self._target(*self._args, **self._kwargs))
+        finally:
+            # Avoid a refcycle if the thread is running a function with
+            # an argument that has a member that points to the thread.
+            del self._target, self._args, self._kwargs
 
 
-def threader(group=None, name=None, daemon=True):
-    """ decorator to thread functions
-    :param group: reserved for future extension when a ThreadGroup class is implemented
-    :param name: thread name
-    :param daemon: thread behavior """
+class DynamicWorker(threading.Thread):
+    """ Thread executing tasks from a given tasks queue
+    :type pool: DynamicPool """
+    def __init__(self, pool, target, args=(), kwargs=None, daemon=None, group=None, name=None):
+        self.pool = pool
+        threading.Thread.__init__(self, group, target, name, args, kwargs, daemon=daemon)
+        self.start()
 
-    def decorator(target):
-        """ :param target: function to be threaded """
+    def run(self):
+        """Method representing the thread's activity.
 
-        def wrapped_job(queue, *args, **kwargs):
-            """ this function calls the decorated function
-            and puts the result in a queue
-            :type queue: Queue """
-            queue.put(target(*args, **kwargs))
+        You may override this method in a subclass. The standard run() method
+        invokes the callable object passed to the object's constructor as the
+        target argument, if any, with sequential and keyword arguments taken
+        from the args and kwargs arguments, respectively.
 
-        def wrap(*args, **kwargs):
-            """ this is the function returned from the decorator. It fires off
-            wrapped_f in a new thread and returns the thread object with
-            the result queue attached
-            :rtype: Thread """
-            thread = Thread(group=group, target=wrapped_job, name=name, args=args, kwargs=kwargs, daemon=daemon)
-            thread.start()
-            return thread
+        """
+        self.pool.threads.append(self)
+        try:
+            if self._target:
+                self.pool.results.put(self._target(*self._args, **self._kwargs))
+        finally:
+            # Avoid a refcycle if the thread is running a function with
+            # an argument that has a member that points to the thread.
+            self.pool.threads.remove(self)
+            del self._target, self._args, self._kwargs
 
-        return wrap
 
-    return decorator
+class ThreadManager(Thread):
+    def __init__(self, pool, group=None, name=None, daemon=True, timeout=0.1):
+        """ Thread executing tasks from a given tasks queue
+        :type pool: DynamicPool """
+        Thread.__init__(self, group=group, name=name, daemon=daemon)
+        self.pool = pool
+        self.start()
+        self.timeout = timeout
+
+    def run(self):
+        while self._is_stopped is False:
+            if self.pool.tasks.qsize() > 0:
+                try:
+                    if self.pool.max_workers is None:
+                        target, args, kwargs = self.pool.tasks.get(timeout=self.timeout)
+                        DynamicWorker(pool=self.pool, target=target, args=args, kwargs=kwargs, daemon=self.pool.daemon)
+                    elif self.pool.max_workers > len(self.pool.threads):
+                        target, args, kwargs = self.pool.tasks.get(timeout=self.timeout)
+                        DynamicWorker(pool=self.pool, target=target, args=args, kwargs=kwargs, daemon=self.pool.daemon)
+                except Empty:
+                    pass
+                except Exception as e:
+                    if self.pool.store_errors is True:
+                        self.pool.errors.put(e)
+                    else:
+                        self.stop()
+                        self.pool.manager = ThreadManager(pool=self.pool, group=None, name=None, daemon=self.daemon,
+                                                          timeout=self.timeout)
+                        del self._target, self._args, self._kwargs
+                        raise e
+
+    def stop(self):
+        self._is_stopped = True
+
+
+class DynamicPool:
+    """Pool of dynamic thread workers consuming tasks from a queue
+    the number of running workers changes by demand"""
+
+    def __init__(self, max_workers=None, daemon=True, timeout=0.1, store_errors=False):
+        """ create a pool of workers, run tasks, collect results """
+        self.threads = []
+        self.results = Queue()
+        self.tasks = Queue()
+        self.errors = Queue()
+        self.daemon = daemon
+        self.timeout = timeout
+        self.max_workers = max_workers
+        self.store_errors = store_errors
+        self.is_stopped = False
+        self.manager = ThreadManager(pool=self, group=None, name=None, daemon=daemon, timeout=timeout)
+
+    def get(self, timeout=None):
+        return self.results.get(timeout)
+
+    def gets(self, timeout=None):
+        if timeout is None:
+            timeout = self.timeout
+        results = []
+        while self.results.qsize() > 0:
+            try:
+                results.append(self.results.get(timeout))
+            except Empty:
+                pass
+        return results
+
+    def put(self, target, *args, **kwargs):
+        """Add a task to the queue"""
+        self.tasks.put((target, args, kwargs))
+
+    def join(self):
+        """Wait for completion of all the tasks in the queue"""
+        if self.is_stopped is False:
+            while not self.tasks.empty():
+                pass
+            self.stop()
+        self.manager.join()
+
+    def stop(self):
+        self.is_stopped = True
+        self.manager.stop()
+
+    def start(self):
+        self.is_stopped = False
+        self.manager.start()
 
 
 class ThreadWorker(Thread):
-    """Thread executing tasks from a given tasks queue
-    :type pool: ThreadPool """
-
     def __init__(self, pool, group=None, name=None, daemon=True, collect_results=False, max_results=0, timeout=0.1):
+        """ Thread executing tasks from a given tasks queue
+        :type pool: ThreadPool """
         Thread.__init__(self, group=group, name=name, daemon=daemon, max_results=max_results)
         self.pool = pool
         self.start()
@@ -102,9 +204,6 @@ class ThreadWorker(Thread):
         self.running_task = None
         self.timeout = timeout
 
-    def start(self):
-        Thread.start(self)
-
     def run(self):
         while self._is_stopped is False:
             if self.pool.lifecycle is not None:
@@ -132,6 +231,7 @@ class ThreadWorker(Thread):
                     else:
                         self.stop()
                         self.pool.add()
+                        del self._target, self._args, self._kwargs
                         raise e
 
     def stop(self):
@@ -139,7 +239,8 @@ class ThreadWorker(Thread):
 
 
 class ThreadPool:
-    """Pool of threads consuming tasks from a queue"""
+    """Pool of thread workers consuming tasks from a queue
+    the number of workers should be static"""
 
     def __init__(
             self,
@@ -209,11 +310,15 @@ class ThreadPool:
 
     def get_all(self, timeout=None):
         results = []
-        while not self.results.empty():
+        while not self.results.empty() or not self.tasks.empty():
             try:
-                results.append(self.results.get(timeout))
-            except:
+                result = self.get(timeout)
+                if result is not None:
+                    results.append(result)
+            except TimeoutError:
                 pass
+        if len(results) == 0:
+            return None
         return results
 
     def get(self, timeout=None):
@@ -299,3 +404,23 @@ def get_first_result(threads, timeout=None):
         if timeout is not None:
             if time() - t >= timeout:
                 raise TimeoutError
+
+
+def threader(group=None, name=None, daemon=True):
+    """ decorator to thread functions
+    :param group: reserved for future extension when a ThreadGroup class is implemented
+    :param name: thread name
+    :param daemon: thread behavior """
+
+    def decorator(target):
+        """ :param target: function to be threaded """
+        def wrap(*args, **kwargs):
+            """ this is the function returned from the decorator. It fires off
+            wrapped_f in a new thread and returns the thread object with
+            the result queue attached
+            :rtype: Thread """
+            thread = Thread(group=group, target=target, name=name, args=args, kwargs=kwargs, daemon=daemon)
+            thread.start()
+            return thread
+        return wrap
+    return decorator
